from typing import Any, Callable, Optional

from pydantic import BaseModel, Field

from prefect._internal.compatibility.async_dispatch import async_dispatch
from prefect._internal.compatibility.migration import getattr_migration
from prefect.client.orchestration import get_client
from prefect.client.schemas.actions import VariableCreate, VariableUpdate
from prefect.client.utilities import get_or_create_client
from prefect.exceptions import ObjectNotFound
from prefect.types import MAX_VARIABLE_NAME_LENGTH, StrictVariableValue


class Variable(BaseModel):
    """
    Variables are named, mutable JSON values that can be shared across tasks and flows.

    Arguments:
        name: A string identifying the variable.
        value: A string that is the value of the variable.
        tags: An optional list of strings to associate with the variable.
    """

    name: str = Field(
        default=...,
        description="The name of the variable",
        examples=["my_variable"],
        max_length=MAX_VARIABLE_NAME_LENGTH,
    )
    value: StrictVariableValue = Field(
        default=...,
        description="The value of the variable",
        examples=["my-value"],
    )
    tags: Optional[list[str]] = Field(default=None)

    @classmethod
    async def aset(
        cls,
        name: str,
        value: StrictVariableValue,
        tags: Optional[list[str]] = None,
        overwrite: bool = False,
    ) -> "Variable":
        """
        Asynchronously sets a new variable. If one exists with the same name, must pass `overwrite=True`

        Returns the newly set variable object.

        Args:
            - name: The name of the variable to set.
            - value: The value of the variable to set.
            - tags: An optional list of strings to associate with the variable.
            - overwrite: Whether to overwrite the variable if it already exists.

        Example:
            Set a new variable and overwrite it if it already exists.

            ```
            from prefect.variables import Variable

            @flow
            async def my_flow():
                await Variable.aset(name="my_var",value="test_value", tags=["hi", "there"], overwrite=True)
            ```
        """
        client, _ = get_or_create_client()
        variable_exists = await client.read_variable_by_name(name)
        var_dict = {"name": name, "value": value, "tags": tags or []}

        if variable_exists:
            if not overwrite:
                raise ValueError(
                    f"Variable {name!r} already exists. Use `overwrite=True` to update it."
                )
            await client.update_variable(
                variable=VariableUpdate.model_validate(var_dict)
            )
            variable = await client.read_variable_by_name(name)
            for key in var_dict.keys():
                var_dict.update({key: getattr(variable, key)})
        else:
            await client.create_variable(
                variable=VariableCreate.model_validate(var_dict)
            )

        return cls.model_validate(var_dict)

    @classmethod
    @async_dispatch(aset)
    def set(
        cls,
        name: str,
        value: StrictVariableValue,
        tags: Optional[list[str]] = None,
        overwrite: bool = False,
    ) -> "Variable":
        """
        Sets a new variable. If one exists with the same name, must pass `overwrite=True`

        Returns the newly set variable object.

        Args:
            - name: The name of the variable to set.
            - value: The value of the variable to set.
            - tags: An optional list of strings to associate with the variable.
            - overwrite: Whether to overwrite the variable if it already exists.

        Example:
            Set a new variable and overwrite it if it already exists.

            ```
            from prefect.variables import Variable

            @flow
            def my_flow():
                Variable.set(name="my_var",value="test_value", tags=["hi", "there"], overwrite=True)
            ```
        """
        with get_client(sync_client=True) as client:
            variable_exists = client.read_variable_by_name(name)
            var_dict = {"name": name, "value": value, "tags": tags or []}

            if variable_exists:
                if not overwrite:
                    raise ValueError(
                        f"Variable {name!r} already exists. Use `overwrite=True` to update it."
                    )
                client.update_variable(variable=VariableUpdate.model_validate(var_dict))
                variable = client.read_variable_by_name(name)
                for key in var_dict.keys():
                    var_dict.update({key: getattr(variable, key)})
            else:
                client.create_variable(variable=VariableCreate.model_validate(var_dict))

            return cls.model_validate(var_dict)

    @classmethod
    async def aget(
        cls,
        name: str,
        default: StrictVariableValue = None,
    ) -> StrictVariableValue:
        """
        Asynchronously get a variable's value by name.

        If the variable does not exist, return the default value.

        Args:
            - name: The name of the variable value to get.
            - default: The default value to return if the variable does not exist.

        Example:
            Get a variable's value by name.
            ```python
            from prefect import flow
            from prefect.variables import Variable

            @flow
            async def my_flow():
                var = await Variable.aget("my_var")
            ```
        """
        client, _ = get_or_create_client()
        variable = await client.read_variable_by_name(name)

        return variable.value if variable else default

    @classmethod
    @async_dispatch(aget)
    def get(
        cls,
        name: str,
        default: StrictVariableValue = None,
    ) -> StrictVariableValue:
        """
        Get a variable's value by name.

        If the variable does not exist, return the default value.

        Args:
            - name: The name of the variable value to get.
            - default: The default value to return if the variable does not exist.

        Example:
            Get a variable's value by name.
            ```python
            from prefect import flow
            from prefect.variables import Variable

            @flow
            def my_flow():
                var = Variable.get("my_var")
            ```
        """
        with get_client(sync_client=True) as client:
            variable = client.read_variable_by_name(name)

            return variable.value if variable else default

    @classmethod
    async def aunset(cls, name: str) -> bool:
        """
        Asynchronously unset a variable by name.

        Args:
            - name: The name of the variable to unset.

        Returns `True` if the variable was deleted, `False` if the variable did not exist.

        Example:
            Unset a variable by name.
            ```python
            from prefect import flow
            from prefect.variables import Variable

            @flow
            async def my_flow():
                await Variable.aunset("my_var")
            ```
        """
        client, _ = get_or_create_client()
        try:
            await client.delete_variable_by_name(name=name)
            return True
        except ObjectNotFound:
            return False

    @classmethod
    @async_dispatch(aunset)
    def unset(cls, name: str) -> bool:
        """
        Unset a variable by name.

        Args:
            - name: The name of the variable to unset.

        Returns `True` if the variable was deleted, `False` if the variable did not exist.

        Example:
            Unset a variable by name.
            ```python
            from prefect import flow
            from prefect.variables import Variable

            @flow
            def my_flow():
                Variable.unset("my_var")
            ```
        """
        with get_client(sync_client=True) as client:
            try:
                client.delete_variable_by_name(name=name)
                return True
            except ObjectNotFound:
                return False


__getattr__: Callable[[str], Any] = getattr_migration(__name__)



from __future__ import annotations

import os
from enum import Enum
from typing import Any, Callable, Coroutine, Dict, Literal, Optional
from urllib.parse import urlparse

from anyio import run_process
from pydantic import Field, model_validator

from prefect.client.schemas.objects import VersionInfo


class SimpleVersionInfo(VersionInfo):
    type: Literal["prefect:simple"] = "prefect:simple"
    version: str = Field(default="")
    branch: Optional[str] = Field(default=None)
    url: Optional[str] = Field(default=None)


class GithubVersionInfo(VersionInfo):
    type: Literal["vcs:github"] = "vcs:github"
    version: str
    branch: str
    url: str
    repository: str

    @model_validator(mode="after")
    def validate_branch(self):
        if not self.branch:
            raise ValueError("branch is required when type is 'vcs:github'")
        return self


class GitVersionInfo(VersionInfo):
    type: Literal["vcs:git"] = "vcs:git"
    version: str
    branch: str
    url: str
    repository: str


async def get_github_version_info(
    version: Optional[str] = None,
    branch: Optional[str] = None,
    repository: Optional[str] = None,
    url: Optional[str] = None,
) -> GithubVersionInfo:
    """Create a GithubVersionInfo object from provided values or environment variables.

    Args:
        version: The commit SHA, falls back to GITHUB_SHA env var
        branch: The git branch, falls back to GITHUB_REF_NAME env var
        repository: The repository name, falls back to GITHUB_REPOSITORY env var
        url: The repository URL, constructed from GITHUB_SERVER_URL/GITHUB_REPOSITORY if not provided

    Returns:
        A GithubVersionInfo

    Raises:
        ValueError: If any required fields cannot be determined
    """
    version = version or os.getenv("GITHUB_SHA")
    branch = branch or os.getenv("GITHUB_REF_NAME")
    repository = repository or os.getenv("GITHUB_REPOSITORY")
    url = url or f"{os.getenv('GITHUB_SERVER_URL')}/{repository}"

    if not version:
        raise ValueError("version is required - must be provided or set in GITHUB_SHA")
    if not branch:
        raise ValueError(
            "branch is required - must be provided or set in GITHUB_REF_NAME"
        )
    if not repository:
        raise ValueError(
            "repository is required - must be provided or set in GITHUB_REPOSITORY"
        )

    return GithubVersionInfo(
        type="vcs:github",
        version=version,
        branch=branch,
        repository=repository,
        url=url,
    )


async def get_git_version_info(
    version: Optional[str] = None,
    branch: Optional[str] = None,
    url: Optional[str] = None,
    repository: Optional[str] = None,
) -> GitVersionInfo:
    try:
        if not version:
            # Run git command and get stdout
            result = await run_process(["git", "rev-parse", "HEAD"])
            # Decode bytes to string and strip whitespace
            version = result.stdout.decode().strip()

        if not branch:
            result = await run_process(["git", "rev-parse", "--abbrev-ref", "HEAD"])
            branch = result.stdout.decode().strip()

        if not repository:
            result = await run_process(["git", "config", "--get", "remote.origin.url"])
            remote_url = result.stdout.decode().strip()

            # Extract just the repository name (last part of the path)
            repo_url = urlparse(remote_url)
            repository = repo_url.path.strip("/")
            if repository.endswith(".git"):
                repository = repository[:-4]

        if not url and repository:
            # Use the full remote URL as the URL
            result = await run_process(["git", "config", "--get", "remote.origin.url"])
            url = result.stdout.decode().strip()
    except Exception as e:
        raise ValueError(
            f"Error getting git version info: {e}. You may not be in a git repository."
        )

    if not url:
        raise ValueError("Could not determine git repository URL")

    return GitVersionInfo(
        type="vcs:git", version=version, branch=branch, url=url, repository=repository
    )


class VersionType(str, Enum):
    SIMPLE = "prefect:simple"
    GITHUB = "vcs:github"
    GIT = "vcs:git"
    DOCKER = "container:docker"


async def get_inferred_version_info(
    version_type: Optional[str] = None,
) -> VersionInfo | None:
    """
    Attempts to infer version information from the environment.

    Args:
        version_type: Optional type of version info to get. If provided, only that
                     type will be attempted.

    Returns:
        VersionInfo: The inferred version information

    Raises:
        ValueError: If unable to infer version info from any source
    """
    # Map version types to their getter functions
    type_to_getter: Dict[str, Callable[..., Coroutine[Any, Any, Any]]] = {
        VersionType.GITHUB: get_github_version_info,
        VersionType.GIT: get_git_version_info,
    }

    # Default order of getters to try
    default_getters = [
        get_github_version_info,
        get_git_version_info,
    ]

    if version_type:
        if version_type not in type_to_getter:
            raise ValueError(f"Unknown version type: {version_type}")
        getters = [type_to_getter[version_type]]
    else:
        getters = default_getters

    for getter in getters:
        try:
            return await getter()
        except ValueError:
            continue

    return None



from __future__ import annotations

import datetime
from datetime import timedelta
from typing import TYPE_CHECKING, Any
from uuid import UUID

import prefect.types._datetime
from prefect.logging.loggers import get_logger

from .clients import (
    AssertingEventsClient,
    AssertingPassthroughEventsClient,
    PrefectCloudEventsClient,
    PrefectEventsClient,
)
from .schemas.events import Event, RelatedResource
from .worker import EventsWorker, should_emit_events

if TYPE_CHECKING:
    import logging

TIGHT_TIMING = timedelta(minutes=5)

logger: "logging.Logger" = get_logger(__name__)


def emit_event(
    event: str,
    resource: dict[str, str],
    occurred: datetime.datetime | None = None,
    related: list[dict[str, str]] | list[RelatedResource] | None = None,
    payload: dict[str, Any] | None = None,
    id: UUID | None = None,
    follows: Event | None = None,
    **kwargs: dict[str, Any] | None,
) -> Event | None:
    """
    Send an event to Prefect.

    Args:
        event: The name of the event that happened.
        resource: The primary Resource this event concerns.
        occurred: When the event happened from the sender's perspective.
                  Defaults to the current datetime.
        related: A list of additional Resources involved in this event.
        payload: An open-ended set of data describing what happened.
        id: The sender-provided identifier for this event. Defaults to a random
            UUID.
        follows: The event that preceded this one. If the preceding event
            happened more than 5 minutes prior to this event the follows
            relationship will not be set.

    Returns:
        The event that was emitted if worker is using a client that emit
        events, otherwise None
    """
    if not should_emit_events():
        return None

    try:
        operational_clients = [
            AssertingPassthroughEventsClient,
            AssertingEventsClient,
            PrefectCloudEventsClient,
            PrefectEventsClient,
        ]
        worker_instance = EventsWorker.instance()

        if worker_instance.client_type not in operational_clients:
            return None

        event_kwargs: dict[str, Any] = {
            "event": event,
            "resource": resource,
            **kwargs,
        }

        if occurred is None:
            occurred = prefect.types._datetime.now("UTC")
        event_kwargs["occurred"] = occurred

        if related is not None:
            event_kwargs["related"] = related

        if payload is not None:
            event_kwargs["payload"] = payload

        if id is not None:
            event_kwargs["id"] = id

        if follows is not None:
            if -TIGHT_TIMING < (occurred - follows.occurred) < TIGHT_TIMING:
                event_kwargs["follows"] = follows.id

        event_obj = Event(**event_kwargs)
        worker_instance.send(event_obj)

        return event_obj
    except Exception:
        logger.exception(f"Error emitting event: {event}")
        return None
